#version 330 core

in vec3 PositionWorldPass;
in vec3 NormalWorldPass;
out vec4 color;

// Values that stay constant for the whole mesh.
uniform mat4 gModel, gView, gProjection;
uniform vec3 gEyePosition; // camera (in world)
uniform vec4 gLightPosition; // light (in world)
uniform vec4 gMaterialColor; // material diffuse color

struct lightSource
{
	vec4 position;
	vec4 ambient;
	vec4 diffuse;
	vec4 specular;
	float constantAttenuation, linearAttenuation, quadraticAttenuation;
	float spotCutoff, spotExponent;
	vec3 spotDirection;
};

lightSource light0 = lightSource(
	vec4(0.2, 0.2, 0.2, 1.0),
	vec4(4.0,  4.0,  4.0, 1.0),
	vec4(1.0,  1.0,  1.0, 1.0),
	vec4(1.0,  1.0,  1.0, 1.0),
	0.0, 1.0, 0.0,
	180.0, 0.0,
	vec3(0.0, 0.0, 0.0)
);

struct material
{
	vec4 ambient;
	vec4 diffuse;
	vec4 specular;
	float shininess;
};

material mymaterial = material(
	vec4(0.1, 0.1, 0.1, 1.0),
	vec4(1.0, 1.0, 1.0, 1.0),
	vec4(1.0, 1.0, 1.0, 1.0),
	5.0
);

void main()
{
	vec3 normalDirection = normalize(NormalWorldPass);
	vec3 viewDirection = normalize(gEyePosition - PositionWorldPass);

	vec3 lightDirection;
	float attenuation;
	float lightPower = 50.0f;

	light0.position = gLightPosition;
	if (light0.position.w == 0.0) // directional light
    {
		attenuation = 1.0; // no attenuation
		lightDirection = normalize(vec3(light0.position));
    }
	else // point or spot light (or other kind of light)
    {
		vec3 vertexToLightSource = light0.position.xyz - PositionWorldPass;
		float distance = length(vertexToLightSource);
		lightDirection = normalize(vertexToLightSource);
		attenuation = 1.0 / (light0.constantAttenuation
                           + light0.linearAttenuation * distance
                           + light0.quadraticAttenuation * distance * distance);
 
		if (light0.spotCutoff <= 90.0) // spotlight
        {
			float clampedCosine = max(0.0, dot(-lightDirection, normalize(light0.spotDirection)));
			if (clampedCosine < cos(radians(light0.spotCutoff))) // outside of spotlight cone
            {
              attenuation = 0.0;
            }
			else
            {
              attenuation = attenuation * pow(clampedCosine, light0.spotExponent);
            }
        }
    }
 
	vec3 ambientLighting = vec3(light0.ambient) * vec3(mymaterial.ambient);
 
 	mymaterial.diffuse = gMaterialColor; 
	vec3 diffuseReflection = attenuation
		* lightPower
		* vec3(light0.diffuse) * vec3(mymaterial.diffuse)
		* max(0.0, dot(normalDirection, lightDirection));
 
	vec3 specularReflection;
	if (dot(normalDirection, lightDirection) < 0.0) // light source on the wrong side?
    {
		specularReflection = vec3(0.0, 0.0, 0.0); // no specular reflection
    }
	else // light source on the right side
    {
		specularReflection = attenuation
			* lightPower
			* vec3(light0.specular) * vec3(mymaterial.specular)
			* pow(max(0.0, dot(reflect(-lightDirection, normalDirection), viewDirection)),
              mymaterial.shininess);
    }
 
	color = vec4(ambientLighting + diffuseReflection + specularReflection, 1.0);
}